package fr.unice.javinspector.core.analyser;

import java.util.*;
import java.lang.reflect.*;


/**
 *
 * <p>Title: JavInspector API for analysing a class</p>
 * <p>Description:API for analysing a class </p>
 * <p>Company: UNSA </p>
 * @author Jerme GAHIDE - Lucas CHARBIT - Gabriel ZERBIB - Xavier GALBOIS
 * @version 1.0
 */
public class ClassifierAnalyser {

  // the inspector of this classifier
  private Inspector inspector;
  // the class analysed
  private Class classAnalysed;
  // the package of class analysed
  private Package classPackage;
  // the modifier of class analysed
  private int classModifiers;
  // the internal classes and interfaces of the class analysed
  private Class[] internalClasses = null;
  // the array of ConstructorAnalyser of this
  private ConstructorAnalyser[] constructor = null;
  // the array of FieldAnalyser of this
  private FieldAnalyser[] field = null;
  // the array of MethodAnalyser of this
  private MethodAnalyser[] method = null;
  // the ArrayList of heritage ClassifierAnaliser
  private ArrayList heritageClassifier = new ArrayList();
  // the ArrayList of all SuperInterfaces
  private ArrayList allSuperInterfaces = new ArrayList();
  // the ArrayList of anonymous internal classes
  private ArrayList anonymousInternalClasses = new ArrayList();
  // the ArrayList of java.lang.Class
  private ArrayList heritageClasses = new ArrayList();

  /**
   * the constructor of this class
   * @param classAnalysed the class to analyse
   * @param inspector the inspector of this
   */
  public ClassifierAnalyser(Class classAnalysed, Inspector inspector) {
    this.classAnalysed = classAnalysed;
    this.inspector = inspector;
    this.classPackage = classAnalysed.getPackage();
    this.classModifiers = classAnalysed.getModifiers();
    this.internalClasses = classAnalysed.getDeclaredClasses();
    // here we deal with constructors
    initialiseConstructor();
    // here we deal with fields
    initialiseField();
    // here we deal with methods
    initialiseMethod();
    // here we deal with the super interface and heritage classes
    if (this.classAnalysed.isInterface()){
      initialiseAllSuperInterfaces(this.classAnalysed);
    }
    else{
      initialiseHeritageClasses();
      initialiseHeritageClassifier();
    }
    initialiseAnonymousInternalClasses();
  }

  /**
   * get the class analysed
   * @return the class analysed by this ClassifierAnalyser
   */
  public Class getClassAnalysed() {
    return this.classAnalysed;
  }

  /**
   * returns the array of ConstructorAnalyser of ALL the constructors
   * @return the array of ConstructorAnalyser of ALL the constructors
   */
  public ConstructorAnalyser[] getConstructor() {
    return constructor;
  }

  /**
   * initialises the informations for each contrustor
   */
  private void initialiseConstructor() {
    Constructor[] cons = classAnalysed.getDeclaredConstructors();
    constructor = new ConstructorAnalyser[cons.length];
    for (int i = 0; i < cons.length; i++) {
      constructor[i] = new ConstructorAnalyser(cons[i],this);
    }
  }

  /**
   * returns the array of FieldAnalyser of ALL fields
   * @return the array of FieldAnalyser of ALL fields
   */
  public FieldAnalyser[] getField() {
    return field;
  }

  /**
   * Initilalise the informations for each field
   */
  public void initialiseField() {
    Field[] fieldTmp = classAnalysed.getDeclaredFields();
    field = new FieldAnalyser[fieldTmp.length];
    for (int i = 0; i < fieldTmp.length; i++) {
      field[i] = new FieldAnalyser(fieldTmp[i],this);
    }
  }

  /**
   * returns an array MethodAnalyzer of ALL Methods
   * @return the array of FieldAnalyser of ALL fields
   */
  public MethodAnalyser[] getMethod() {
    return method;
  }

  /**
   * Initialise the informations for each Method
   */
  private void initialiseMethod() {
    Method[] methodTmp = classAnalysed.getDeclaredMethods();
    method = new MethodAnalyser[methodTmp.length];
    for (int i = 0; i < methodTmp.length; i++) {
      method[i] = new MethodAnalyser(methodTmp[i],this);
    }

  }

  /**
   * initialise the arraylist of all the superclasses of the analysed class
   */
  public void initialiseHeritageClassifier() {
      heritageClassifier = new ArrayList();
      heritageClassifier.add(this);
      if (!classAnalysed.isInterface()) {
        Class currentClass = classAnalysed;
        Class currentParentClass = null;
        while ( (currentParentClass = currentClass.getSuperclass()) != null) {
          ClassifierAnalyser classAn =  inspector.getClassifierAnalyser(currentParentClass);
          heritageClassifier.add(classAn);
          currentClass = currentParentClass;
        }
      }
    }

    /**
     * this function initialise the heritage classes
     */
    public void initialiseHeritageClasses() {
      heritageClasses.add(classAnalysed);
      if (!classAnalysed.isInterface()) {
        Class currentClass = classAnalysed;
        Class currentParentClass = null;
        while ( (currentParentClass = currentClass.getSuperclass()) != null) {
          heritageClasses.add(currentParentClass);
          currentClass = currentParentClass;
        }
      }
    }

    /**
     * get the list of heritage ClassifierAnalyser
     * @return the list of heritage ClassifierAnalyser
     */
    public ArrayList getHeritageClassifier(){
      return heritageClassifier;
    }

  /**
   * get the class modifiers of the analysed class(attributes)
   * @return the class modifiers of this class(attributes)
   */
  public int getClassModifiers() {
    return classModifiers;
  }

  /**
   * get the package of the analysed class
   * @return the package of the analysed class
   */
  public Package getClassPackage() {
    return classPackage;
  }

  /**
   * parses the name of the class <code>c</code> to find its type and its realname
   * (case the class is an array)
   * @param c the class to parse
   * @return all the information abtained
   */
  public TypeClass formatParametre(Class c) {
    String s = c.getName();
    if (s.charAt(0) == '[') {
      StringBuffer param = new StringBuffer("");
      int dimension = 0;
      while (s.charAt(dimension) == '[')
        dimension++;
      switch (s.charAt(dimension)) {
        case 'B':
          return createTypeClass(dimension, Byte.TYPE);
        case 'C':
          return createTypeClass(dimension, Character.TYPE);
        case 'D':
          return createTypeClass(dimension, Double.TYPE);
        case 'F':
          return createTypeClass(dimension, Float.TYPE);
        case 'I':
          return createTypeClass(dimension, Integer.TYPE);
        case 'J':
          return createTypeClass(dimension, Long.TYPE);
        case 'S':
          return createTypeClass(dimension, Short.TYPE);
        case 'Z':
          return createTypeClass(dimension, Boolean.TYPE);
        case 'L':
          try {
	    Class tmp = inspector.getMainClassLoader().
               loadClass(param.append(s.substring(dimension + 1,
						  s.indexOf(";"))).toString());
            return createTypeClass(dimension, tmp);
          }
          catch (Exception e) {
            e.printStackTrace(System.out);
          }
      }
      return null;
    }
    else
      return createTypeClass(0, c);
  }

  /**
   * creates a TypeClass object from the dimension and the name of the class
   * @param dimension the diamension of the array (0 if not an array)
   * @param cl the class
   * @return the TypeClass object containing the real understandable name of
   * the analysed class
   */
  public static TypeClass createTypeClass(int dimension, Class cl) {
    String tmp = "";
    for (int i = 0; i < dimension; i++) {
      tmp += "[]";
    }
    return new TypeClass(cl.getName() + tmp, cl, dimension);
  }

  /**
   * this function returns the short name of a class name
   * if the name is java.lang.String this function return String
   * @param s the name of class
   * @return the short name of s
   */
  public static String shortName(String s) {
    StringTokenizer st = new StringTokenizer(s, ".");
    String tmp = null;
    while (st.hasMoreTokens()) {
      tmp = st.nextToken();
    }
    return tmp;
  }

  /**
   * this function returns the list of all the exceptions
   * throwed by the methods
   * @return the list of all the exceptions
   */
  public ArrayList getMethodsExceptions() {
    ArrayList l = new ArrayList();
    for (int i = 0; i < method.length; i++) {
      Class[] tmp = method[i].getMethod().getExceptionTypes();
      for (int k = 0; k < tmp.length; k++) {
        if (l.indexOf(tmp[k]) == -1) {
          l.add(tmp[k]);
        }
      }
    }
    return l;
  }

  /**
   * this function return the ArrayList with all the classes used by the analysed class
   * @return the ArrayList containing java.lang.Class with all the classes used by the analysed class
   */
  public ArrayList getConstructorsParameterType() {
    ArrayList l = new ArrayList();
    for (int i = 0; i < constructor.length; i++) {
      TypeClass[] tmp = constructor[i].getListTypeClass();
      for (int j = 0; j < tmp.length; j++) {
        if (l.indexOf(tmp[j].getCl()) == -1) {
          l.add(tmp[j].getCl());
        }
      }
    }
    return l;
  }

  /**
   * get all the exceptions thrown by any contructors
   * @return an arraylist of java.lang.Class object describing the exceptions
   */
  public ArrayList getConstructorsException() {
    ArrayList l = new ArrayList();
    for (int i = 0; i < constructor.length; i++) {
      Class[] tmp = constructor[i].getConstructor().getExceptionTypes();
      for (int j = 0; j < tmp.length; j++) {
        if (l.indexOf(tmp[j]) == -1) {
          l.add(tmp[j]);
        }
      }
    }
    return l;
  }

  /**
   * get all the type of class for all the fields
   * @return arraylist containing <code>Class</code> of the view
   */
  public ArrayList getTypeField() {
    ArrayList l = new ArrayList();
    for (int i = 0; i < field.length; i++) {
      Class tmp = field[i].getTypeClass().getCl();
      if (l.indexOf(tmp) == -1) {
        l.add(tmp);
      }
    }
    return l;
  }

  /**
   * this function returns all the type of parameters of all methods
   * @return an ArrayList of java.lang.Class
   */
  public ArrayList getMethodsParameterType() {
    ArrayList l = new ArrayList();
    for (int i = 0; i < method.length; i++) {
      TypeClass[] tmp = method[i].getListTypeClass();
      for (int j = 0; j < tmp.length; j++) {
        if (l.indexOf(tmp[j].getCl()) == -1) {
          l.add(tmp[j].getCl());
        }
      }
    }
    return l;
  }

  /**
   * this function returns all the type of return type of all methods
   * @return an ArrayList of java.lang.Class
   */
  public ArrayList getMethodReturnType() {
    ArrayList l = new ArrayList();
    for (int i = 0; i < method.length; i++) {
      Class tmp = method[i].getReturnType().getCl();
      if (l.indexOf(tmp) == -1) {
        l.add(tmp);
      }
    }
    return l;
  }

  /**
   * initialise all the super interfaces of the class cl
   * @param cl the class to deal with
   */
  private void initialiseAllSuperInterfaces(Class cl) {
    if (cl.getInterfaces().length == 0);
    else {
      for (int i = 0; i < cl.getInterfaces().length; i++) {
        if (allSuperInterfaces.indexOf(cl.getInterfaces()[i]) == -1)
          allSuperInterfaces.add(cl.getInterfaces()[i]);
        initialiseAllSuperInterfaces(cl.getInterfaces()[i]);
      }
    }
  }

  /**
   * get all the super interfaces of the class analysed
   * @return the list of java.lang.Class of super interfaces of the class analysed
   */
  public ArrayList getAllSuperInterfaces() {
    return allSuperInterfaces;
  }

  /**
   * this function return an Array of java.lang.Class of all the
   * SuperInterfaces
   * @return an Array of java.lang.Class of all the SuperInterfaces
   */
  public ClassifierAnalyser[] getAllSuperInterfacesClassifier() {
     ClassifierAnalyser [] res = new ClassifierAnalyser[allSuperInterfaces.size()];
     for(int i = 0;i<allSuperInterfaces.size();i++){
       ClassifierAnalyser classAn =  inspector.
           getClassifierAnalyser((Class)allSuperInterfaces.get(i));
       res[i] = classAn;
     }
     return res;
   }

  /**
   * get the internal classes declared by this class
   * @return the array containing all the internal classes
   */
  public Class[] getInternalClasses() {
    if (internalClasses == null) {
      internalClasses = classAnalysed.getDeclaredClasses();
    }
    return internalClasses;
  }

  /**
   * get the name of an internal class
   * @param cl an internal class
   * @return the name of the internal class
   */
  public static String nameInternalClass(Class cl) {
    StringTokenizer st = new StringTokenizer(cl.getName(), "$");
    String res = "";
    while (st.hasMoreTokens()) {
      res = st.nextToken();
    }
    return res;
  }

  /**
   * get the modifiers(attributes) of the analysed class in a String
   * @param cl the class
   * @return the String containing the modifiers
   */
  public static String modifierToString(Class cl) {
    int modifier = cl.getModifiers();
    return Modifier.toString(modifier);
  }

  /**
   * Determines the interfaces implemented by the class represented by this object.
   * @return an array of ClassifierAnalyser
   */
  public ClassifierAnalyser[] getInterfacesClassifier(){
    ClassifierAnalyser [] res = new ClassifierAnalyser[classAnalysed.
        getInterfaces().length];
    for (int i = 0; i < classAnalysed.getInterfaces().length; i++) {
      res[i]=  inspector.getClassifierAnalyser(classAnalysed.getInterfaces()[i]);
    }
    return res;
  }

  /**
   * get the modifier in String
   * @return a String that represents the modifiers
   */
  public String modifierToString() {
    return Modifier.toString(classAnalysed.getModifiers());
  }

  public String getCompleteName() {
    if (!classAnalysed.isInterface()) {
      return "class " + classAnalysed.getName();
    }
    else
      return classAnalysed.getName();
  }

  /**
   * this function returns the full name of the classifier
   * @return a String which represents the full name of the classifier
   */
  public String fullName() {
    String clas = " class ";
    String extend = " extends ";
    if (classAnalysed.isInterface())
      clas = " ";
    return modifierToString() + clas + classAnalysed.getName();
  }

  /**
   * get the name of a ClassifierAnanlyser
   * @return the full name of the classifier
   */
  public String toString() {
    String res = this.modifierToString();
    res += " " + classAnalysed.toString();
    if (classAnalysed.isInterface()) {
      if (classAnalysed.getInterfaces().length > 0) {
        res += " extends ";
        int i = 0;
        for (i = 0; i < classAnalysed.getInterfaces().length - 1; i++) {
          res += classAnalysed.getInterfaces()[i].getName() + ", ";
        }
        res += classAnalysed.getInterfaces()[i].getName();
      }
    }
    else {
      if (classAnalysed.getSuperclass() != null) {
        res += " extends ";
        res += classAnalysed.getSuperclass().getName();
      }
      if (classAnalysed.getInterfaces().length > 0) {
        res += " implements ";
        int i = 0;
        for (i = 0; i < classAnalysed.getInterfaces().length - 1; i++) {
          res += classAnalysed.getInterfaces()[i].getName() + ", ";
        }
        res += classAnalysed.getInterfaces()[i].getName();
      }
    }
    return res;
  }

  /**
   * this function initialise the anonym internal classes.
   */
  private void initialiseAnonymousInternalClasses() {
  	ClassLoader cl = classAnalysed.getClassLoader();
  	if (cl != null){
    int i = 1;
    while (true) {
      try {
        Class c = cl.loadClass(classAnalysed.
            getName() + "$" + i);
        anonymousInternalClasses.add(c);
        i++;
      }
      catch (ClassNotFoundException e) {
      		System.out.println("not Found"+i+classAnalysed.
            getName());
            System.out.println(e);
        break;
      }
      catch (SecurityException s) {
      	System.out.println("securite"+i+classAnalysed.
            getName());
            System.out.println(s);
        break;
      }
      catch (Exception a) {
      	System.out.println("EXCEPEtion"+i+classAnalysed.
            getName());
            System.out.println(a);
        break;
      }
      }
    }
  }

  /**
   * get a list of java.lang.Class of all the internal anonymous classes
   * @return reurns all the anonymous classes in this classifier
   */
  public ArrayList getAnonymousInternalClasses() {
    return anonymousInternalClasses;
  }

  /**
   * this function returns all the methods in the classifier
   * who are already defined in the class cl
   * @param cl the class where we search the methods
   * @return an ArrayList that contains all the methods
   */
  public ArrayList getMethodRedefinedInClass(Class cl) {
    Method m = null;
    ArrayList res = new ArrayList();
    for (int j = 0; j < method.length; j++) {
      try {
        Method comp = null;
        Method meth = method[j].getMethod();
        if ( (comp = getMethod(meth.getName(), meth.getParameterTypes(), cl,
                               meth.getReturnType(), meth.getModifiers())) != null) {
          res.add(comp);
        }
      }
      catch (Exception e) {
        System.err.println("error in method getMethodRedefinedInClass in class" +
                           " ClassifierAnalyser");
        e.printStackTrace();
      }
    }
    return res;
  }

  /**
   * return a method that matchs with name, parameterTypes, returnType, modifier
   * and which is in cl
   * @param name the name of method
   * @param parameterTypes the parameters of method
   * @param cl the java.lang.reflect.Class
   * @param returnType the return type of method
   * @param modifier the modofier of method
   * @return the method that matchs with name, parameterTypes, cl, returnType, modifier
   * or null if there is no method
   */
  public Method getMethod(String name, Class[] parameterTypes, Class cl,
                          Class returnType, int modifier) {
    Method[] meth = cl.getDeclaredMethods();
    for (int i = 0; i < meth.length; i++) {
      if (meth[i].getName().compareTo(name) == 0) {
        if (meth[i].getModifiers() == modifier) {
          if (meth[i].getReturnType() == returnType) {
            Class[] tmp = meth[i].getParameterTypes();
            if ( (tmp.length == parameterTypes.length) && (tmp.length == 0)) {
              return meth[i];
            }
            if (tmp.length == parameterTypes.length) {
              for (int j = 0; j < parameterTypes.length; j++) {
                if (parameterTypes[j] != tmp[j]) {
                  return null;
                }
                else;
              }
              return meth[i];
            }
            else;
          }
        }
      }
      else;
    }
    return null;
  }

  /**
   * this function return all the super classes of the class analysed
   * @return a list of java.lang.Class that are all the super classes of the class analysed
   */
  public ArrayList getHeritageClasses() {
    return heritageClasses;
  }

}
