package fr.unice.javinspector.core.engineSearch;

import java.util.ArrayList;
import java.lang.reflect.Member;

/**
 * <p>Title: Criterion</p>
 * <p>Description: This interface allow to make some new criterion to
 * complete the search engine.
 * </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: UNICE</p>
 * @author Zerbib Gabriel, Charbit Lucas, Galbois Xavier Gahide Jrme
 * @version 1.0
 */

public abstract class Criterion{

  /**
   * The name of this criterion
   * By default, this is the string undefined. To be override with the name
   * of the class that extends this one.
   */
  protected String name = "Undefined";

  /**
   * This is the list of all uncompatible criterion with this criterion
   * By default, init to empty.
   */
  protected ArrayList uncompatibleCriterionList = new ArrayList();

  /**
   * This is the type of activation of the criterion.
   * By default, init at DONTMIND.
   * @see EngineSearchConst
   */
  public int isActive = EngineSearchConst.DONTMIND;

  /**
   * This represent the logical link (AND,OR) between this criterion
   * and the previous criterion.
   * By default, init at AND.
   * @see EngineSearchConst
   */
  protected int logicalLink = EngineSearchConst.AND;

  /**
   * An arrayList of additionnal component for this criterion if needed
   * This is added for some front end need.
   */
  protected ArrayList AdditionalComponent = new ArrayList();


  /**
   * @return the name of the criterion
   */
  public String getName(){
    return name;
  }

  /**
   * set the status of the criterion ( YES, NOT, DONTMIND)
   * @param on_off the status of the criterion.
   * @throws SearchEngineException
   * @see EngineSearchConst
   */
  public void setUseCriterion(int on_off)
      throws SearchEngineException{

    if(on_off != EngineSearchConst.YES &&
       on_off != EngineSearchConst.NOT &&
       on_off != EngineSearchConst.DONTMIND  ){
      throw new SearchEngineException("setUseCriterion : invalid parameters :"+on_off);
    }

    this.isActive = on_off;
  }

  /**
   * @return the integer that tells the status of the criterion.
   */
  public int isCriterionUseable(){
    return isActive;
  }

  /**
   * @return the integer that tells the type of link of the criterion.
   */
  public int getLogicalLink(){
    return logicalLink;
  }

  /**
   * Change the logical link of the criterion
   * This is very usefull in the EngineSearch.
   * @param ll the new link type ( AND, OR)
   * @throws SearchEngineException
   * @see EngineSearchConst
   * @see EngineSearch
   */
  public void setLogicalLink(int ll)
      throws  SearchEngineException{
    if(ll != EngineSearchConst.AND &&
       ll != EngineSearchConst.OR ){
      throw new SearchEngineException("setLogicalLink : invalid parameters :"+ ll );
    }
    this.logicalLink = ll;
  }

  /**
   * Add a criterion that is not compatible with this criterion.
   * @param c the uncompatible criterion
   * @deprecated you can use it, but it will have no effect because the manager
   * is not implemented yet in the search.
   */
  public void addNonCompatibilityCriterion(Criterion c){
    uncompatibleCriterionList.add(c);
  }

  /**
   * return the list of all the criterion that are not compatible
   * whith this one. It's an arrayList of String.
   * @return the ArrayList of String.
   * @deprecated It was to use with a manager of non compatible
   * criterion, but it is not implemented.
   */
  public ArrayList getNonCompatibilityCriterion(){
    return uncompatibleCriterionList;
  }

  /**
   * To implemented for each Criterion
   * @param m the member that we want to check the validity.
   * @return true if the Criterion is Valid for this Member
   * @throws SearchEngineException
   */
  public abstract boolean isCriterionValid(Member m)
      throws SearchEngineException;


  /**
   *
   * @return Empty if there no additionnal graphics component for a criterion
   * @since 2.0
   */
  public ArrayList isAdditionnalGraphicsComponent(){
    return AdditionalComponent;
  }


}
